// C/C++ Headers

#include <stdio.h>
#include <stdlib.h>

// Our headers

#include "QComponent.h"
#include "cpp_assert.h"

// Qt Headers

#include <qwidget.h>
#include <qrect.h>
#include <qobject.h>
#include <qvariant.h>

/*****************************************************************************
 * Definition of class: QComponent
 *
 * This class records each move/resize command and uses that information
 * to report back to callers about its location and dimensions.
 * We cannot rely on the underlying Component object for size/location
 * information as its values are not updated immediately after being
 * set.
 *****************************************************************************/

// Initialize class-level variable 'mVerboseOutput'

enum { UNSPECIFIED = -1, YES, NO };
int QComponent::mVerboseOutput = UNSPECIFIED;

QComponent::QComponent(int index, QWidget *q)
   : Component(index, q->sizeHint().width(), q->sizeHint().height()),
     mWidget(q),
     mWidth((double)Component::getDefaultWidth()),
     mHeight((double)Component::getDefaultHeight()),
     mXLocation(0.0),
     mYLocation(0.0)
{
   if(mVerboseOutput == UNSPECIFIED)
      mVerboseOutput = (getenv(QALE_VERBOSE) ? YES : NO);
}

QComponent::~QComponent()
{
   // Nothing to do
}

double QComponent::getLeft() const
{
   double x, y;

   getLocation(&x, &y);
   return(x);
}

double QComponent::getRight() const
{
   double x, y, width, height;

   getSize(&width, &height);
   getLocation(&x, &y);

   return(x + width);
}

double QComponent::getTop() const
{
   double x, y;

   getLocation(&x, &y);

   return(y);
}

double QComponent::getBottom() const
{
   double x, y, width, height;

   getSize(&width, &height);
   getLocation(&x, &y);

   return(y + height);
}

double QComponent::getX() const
{
   double x, y, width, height;

   getSize(&width, &height);
   getLocation(&x, &y);

   return(x + (width / 2.0));
}

double QComponent::getY() const
{
   double x, y, width, height;

   getSize(&width, &height);
   getLocation(&x, &y);

   return(y + (height / 2.0));
}

double QComponent::getWidth() const
{
   double width, height;

   getSize(&width, &height);

   return(width);
}

double QComponent::getHeight() const
{
   double width, height;

   getSize(&width, &height);

   return(height);
}

void QComponent::setLocation(double x, double y)
{
   mXLocation = x;
   mYLocation = y;

   int xValue = (int)(x + 0.5);
   int yValue = (int)(y + 0.5);

   if(mVerboseOutput == YES)
   {
      printf(
         "%s->move(%d, %d)\n",
         mWidget->objectName().toLatin1().data(), xValue, yValue
      );
   }

   mWidget->move(xValue, yValue);
}

void QComponent::setSize(double width, double height)
{
   mWidth  = width;
   mHeight = height;

   int widthValue  = (int)(width + 0.5);
   int heightValue = (int)(height + 0.5);

   if(mVerboseOutput == YES)
   {
      printf(
         "%s->resize(%d, %d)\n",
         mWidget->objectName().toLatin1().data(), widthValue, heightValue
      );
   }

   mWidget->resize(widthValue, heightValue);
}

int QComponent::hasProperty(const char *propertyName) const
{
   return(mWidget->property(propertyName).isValid());
}

double QComponent::getProperty(const char *propertyName) const
{
   QVariant value = mWidget->property(propertyName);

   CPP_ASSERT(value.isValid());
   return(value.toDouble());
}

void QComponent::getSize(double *widthPtr, double *heightPtr) const
{
   *widthPtr  = mWidth;
   *heightPtr = mHeight;
}

void QComponent::getLocation(double *xPtr, double *yPtr) const
{
   *xPtr = mXLocation;
   *yPtr = mYLocation;
}

QWidget *QComponent::getWidget() const
{
   return(mWidget);
}
