%{

#include <iostream>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <sstream>
#include <string>

using std::string;
using std::ostringstream;

// Include malloc.h to satisfy reference to "alloca" in bison-generated
// code on win32 platforms

#ifdef _WIN32
#   include <malloc.h>
#endif

#include "QALELexer.h"

#include "Node.h"
#include "BinaryNode.h"
#include "UnaryNode.h"
#include "ConstantNode.h"
#include "PropertyNode.h"
#include "AttributeNode.h"
#include "DefineNode.h"
#include "MinMaxNode.h"
#include "Alignment.h"
#include "StackTemplate.h"
#include "Widget.h"
#include "Define.h"
#include "PointerSet.h"

#include "qale_parse.h"

#define YYERROR_VERBOSE (1)
#define YYDEBUG         (0)
#define ENV(x)          ((QALEParse *)voidPointer)->x
#define YYLEX_PARAM     &(ENV(theIdentifierSet))
#define YYPARSE_PARAM   voidPointer

// A little trickiness to allow an additional variable to be passed
// to qale_yyerror when it is called - this avoids the need for
// static (or global) data to record qale_yyerror's error messages

#define qale_yyerror(message) handleError(message, voidPointer)

extern "C" {
   int yylex(...);
}

// Typedefs for the template stuff we'll need to use
// inside QALEParse

typedef StackTemplate<Widget>      WidgetStack;
typedef StackTemplate<WidgetStack> WidgetStackStack;

typedef StackTemplate<Define>      DefineStack;
typedef StackTemplate<DefineStack> DefineStackStack;

// Local class representing parsing environment
// We need a prototype for the following function since its used
// as an argument to the constructor for "theNodeSet"

static void nodeDelete(void *ptr);

class QALEParse
{
   public:
      ostringstream errStream;
      int errorLineNumber;
      WidgetStackStack theWidgets;
      DefineStackStack theDefines;
      PointerSet theIdentifierSet;
      PointerSet theNodeSet;

      QALEParse()
         :
         errorLineNumber(0),
         theIdentifierSet(PointerSet::genericArrayDelete),
         theNodeSet(nodeDelete)
      {
      }
};

/*****************************************************************************
 * Local function prototypes
 *****************************************************************************/

static int handleError(char *s, void *voidPointer);

static Widget *qale_createFromText(
   const char *text,
   QALEParse  *env,
   string     &errMsg,
   int *errorLineNumberPointer
);

%}

%pure_parser

%union
{
   double                   dvalue;
   char                    *identifier;
   char                    *qualified_identifier;
   AttributeNode::Attribute theAttribute;
   Alignment::Type          theType;
   Node                    *n;
}

// Tokens ...

%token <dvalue>               REAL
%token <identifier>           IDENTIFIER
%token <qualified_identifier> QUALIFIED_IDENTIFIER

%token <theAttribute> LEFT
%token <theAttribute> RIGHT
%token <theAttribute> TOP
%token <theAttribute> BOTTOM
%token <theAttribute> XCENTER
%token <theAttribute> YCENTER
%token <theAttribute> WIDTH
%token <theAttribute> HEIGHT
%token <theAttribute> DEFAULT_WIDTH
%token <theAttribute> DEFAULT_HEIGHT

%token <theType>      LT
%token <theType>      LC
%token <theType>      LB
%token <theType>      CT
%token <theType>      CC
%token <theType>      CB
%token <theType>      RT
%token <theType>      RC
%token <theType>      RB

%token DEFINE
%token MIN
%token MAX

// Types for productions that needs them

%type <n>             expression
%type <n>             property_expression
%type <n>             variable_expression
%type <n>             attribute_expression
%type <theAttribute>  the_attribute
%type <theType>       the_alignment

// Handle precedence

%left '-' '+'
%left '*' '/'
%nonassoc UMINUS

%%

// Main production

widget: IDENTIFIER '('
          expression ','
          expression ','
          expression ','
          expression ','
          the_alignment
      ')'
       {
          // A widget may or may not have a "{ ... }" section
          // create stacks that we can examine after we've dealt
          // with the "widget_contents" production to see if
          // this widget did have contents

          ENV(theWidgets).push(new WidgetStack());
          ENV(theDefines).push(new DefineStack());
       }
       widget_contents
       {
          WidgetStack *children = ENV(theWidgets).pop();
          DefineStack *defines  = ENV(theDefines).pop();
          int level             = ENV(theWidgets).size() - 1;

         ENV(theNodeSet).remove($3);
         ENV(theNodeSet).remove($5);
         ENV(theNodeSet).remove($7);
         ENV(theNodeSet).remove($9);

          Widget *w = new Widget(
             $1, $3, $5, $7, $9,
             new Alignment($11),
             level
          );

          ENV(theIdentifierSet).remove($1);
          delete [] $1;

          int nDefines  = defines->size();

          if(nDefines > 0)
          {
             w->setNDefines(nDefines);

             // Add the defines to our widget, note that we
             // correct the order which will have been reversed
             // by virtue of having stored them in a stack

             for(int i = nDefines - 1; i >= 0; i --)
                w->setDefine(i, defines->pop());
          }

          delete defines;

          int nChildren = children->size();

          if(nChildren > 0)
          {
             w->setNChildren(nChildren);

             // Add the children to our widget, note that we
             // correct the order which will have been reversed
             // by virtue of having stored them in a stack

             for(int i = nChildren - 1; i >= 0; i --)

                w->setChild(i, children->pop());
          }

          delete children;

          WidgetStack *ws = ENV(theWidgets).top();
          ws->push(w);
       }
   ;

widget_contents: /* -- empty, a widget might not have contents -- */
   |             '{' define_contents children '}'
   ;

children: widget
   |      children widget
   ;

define_contents: /* -- empty, a widget might not have any defines -- */
   |             DEFINE '{' defines '}'
   ;

defines: define
   |     defines define
   ;

define: IDENTIFIER '=' expression
         {
            const char *name  = $1;
            Node *value       = $3;
            int level         = ENV(theWidgets).size() - 1;

            ENV(theNodeSet).remove(value);

            Define *d = new Define(name, value, level);
            ENV(theIdentifierSet).remove($1);
            delete [] $1;

            DefineStack *ds = ENV(theDefines).top();
            ds->push(d);
         }
   ;

expression: REAL
            {
               $$ = new ConstantNode($1);
               ENV(theNodeSet).add($$);
            }
   |        expression '+' expression
            {
               ENV(theNodeSet).remove($1);
               ENV(theNodeSet).remove($3);

               $$ = new BinaryNode($1, BinaryNode::_PLUS_, $3);

               ENV(theNodeSet).add($$);
            }
   |        expression '-' expression
            {
               ENV(theNodeSet).remove($1);
               ENV(theNodeSet).remove($3);

               $$ = new BinaryNode($1, BinaryNode::_MINUS_, $3);

               ENV(theNodeSet).add($$);
            }
   |        expression '*' expression
            {
               ENV(theNodeSet).remove($1);
               ENV(theNodeSet).remove($3);

               $$ = new BinaryNode($1, BinaryNode::_TIMES_, $3);

               ENV(theNodeSet).add($$);
            }
   |        expression '/' expression
            {
               ENV(theNodeSet).remove($1);
               ENV(theNodeSet).remove($3);

               $$ = new BinaryNode($1, BinaryNode::_DIVIDE_, $3);

               ENV(theNodeSet).add($$);
            }
   |        '-' expression %prec UMINUS
            {
               ENV(theNodeSet).remove($2);

               $$ = new UnaryNode($2);

               ENV(theNodeSet).add($$);
            }
   |        '(' expression ')'
            {
               $$ = $2;
            }
   |        MIN '(' expression ',' expression ')'
            {
               ENV(theNodeSet).remove($3);
               ENV(theNodeSet).remove($5);

               $$ = new MinMaxNode(MinMaxNode::_MIN_, $3, $5);

               ENV(theNodeSet).add($$);
            }
   |        MAX '(' expression ',' expression ')'
            {
               ENV(theNodeSet).remove($3);
               ENV(theNodeSet).remove($5);

               $$ = new MinMaxNode(MinMaxNode::_MAX_, $3, $5);

               ENV(theNodeSet).add($$);
            }
   |        property_expression
            {
               $$ = $1;
            }
   |        variable_expression
            {
               $$ = $1;
            }
   |        attribute_expression
            {
               $$ = $1;
            }
   ;

property_expression: QUALIFIED_IDENTIFIER '(' ')'
                     {
                        $$ = new PropertyNode($1);
                        ENV(theIdentifierSet).remove($1);
                        delete [] $1;

                     }
   |               IDENTIFIER '(' ')'
                     {
                        $$ = new PropertyNode($1);
                        ENV(theIdentifierSet).remove($1);
                        delete [] $1;
                     }
   ;

variable_expression: '$' QUALIFIED_IDENTIFIER
                     {
                        $$ = new DefineNode($2);
                        ENV(theNodeSet).add($$);

                        ENV(theIdentifierSet).remove($2);
                        delete [] $2;
                     }
   |                 '$' IDENTIFIER
                     {
                        $$ = new DefineNode($2);
                        ENV(theNodeSet).add($$);

                        ENV(theIdentifierSet).remove($2);
                        delete [] $2;
                     }
   ;

attribute_expression: QUALIFIED_IDENTIFIER
                        {
                           string errMsg;

                           AttributeNode *ptr =
                              AttributeNode::createFromQualifiedIdentifier(
                                 $1, errMsg
                              );

                           ENV(theIdentifierSet).remove($1);
                           delete [] $1;

                           if(ptr == NULL)
                           {
                              handleError(
                                 (char *)errMsg.c_str(),
                                 voidPointer
                              );
                              return(-1);
                           }

                           $$ = ptr;
                           ENV(theNodeSet).add($$);
                        }
   |                  the_attribute
                        {
                           $$ = new AttributeNode("", $1);
                           ENV(theNodeSet).add($$);
                        }
   ;

the_attribute: LEFT           { $$ = $1; }
   |           RIGHT          { $$ = $1; }
   |           TOP            { $$ = $1; }
   |           BOTTOM         { $$ = $1; }
   |           XCENTER        { $$ = $1; }
   |           YCENTER        { $$ = $1; }
   |           WIDTH          { $$ = $1; }
   |           HEIGHT         { $$ = $1; }
   |           DEFAULT_WIDTH  { $$ = $1; }
   |           DEFAULT_HEIGHT { $$ = $1; }
   ;

the_alignment: LT { $$ = $1; }
   |           LC { $$ = $1; }
   |           LB { $$ = $1; }
   |           CT { $$ = $1; }
   |           CC { $$ = $1; }
   |           CB { $$ = $1; }
   |           RT { $$ = $1; }
   |           RC { $$ = $1; }
   |           RB { $$ = $1; }
   ;

%%

/*****************************************************************************
 * Local functions
 *****************************************************************************/

static int handleError(char *s, void *voidPointer)
{
   char currentLine[QALELexer::MAX_LINE_LENGTH + 1];
   QALELexer *lex = QALELexer::getInstance();

   // Record the line number where the error took place

   ENV(errorLineNumber) = lex->getLineNumber();

   lex->getCurrentLine(currentLine);

   ENV(errStream) << "line " << ENV(errorLineNumber) << ": " << s << "\n";
   ENV(errStream) << currentLine << "\n";

   int lineOffset = lex->getLineOffset();

   // Attempt to point out the place in the most recently processed
   // line where the error occurred

   for(int i = 0; i < lineOffset; i ++)
   {
      ENV(errStream) << ' ';
   }

   ENV(errStream) << '^' << '\0';

   return(0);
}

static Widget *qale_createFromText(
   const char *text,
   QALEParse  *env,
   string     &errMsg,
   int        *errorLineNumberPointer
)
{
#if YYDEBUG != 0
   yydebug = 1;
#endif

   errMsg                  = "";
   *errorLineNumberPointer = 0;

   // Create our top-level WidgetStack ...

   env->theWidgets.push(new WidgetStack());

   QALELexer::getInstance()->begin(text);
   int success = (qale_yyparse(env) == 0);
   QALELexer::getInstance()->end();

   Widget *w = 0;

   if(!success)
   {
      // Ensure any error messages make their way to the client, also
      // grab the line number of the error

      errMsg                  = env->errStream.str();
      *errorLineNumberPointer = env->errorLineNumber;
   }
   else
   {
      // The data stream was ok syntactically but we need to check a few
      // additional things:
      //
      //    (a) We should have had exactly one top-level widget, meaning:
      //        - theWidgets should contain exactly one element
      //        - the one element in theWidgets should itself contain
      //          exactly one element
      //
      //    (b) "theDefines" should be empty
      //
      // These are really just "sanity checks", they shouldn't ever actually
      // fail - but check anyway

      if(env->theDefines.size() != 0)
      {
         ostringstream errStream;

         errStream << "Expected zero items in \"theDefines\", found: ["
                   << env->theDefines.size() << "]";

         errMsg = errStream.str();
      }
      else
      {
         if(env->theWidgets.size() != 1)
         {
            ostringstream errStream;

            errStream << "Expected one item in \"theWidgets\", found: ["
                      << env->theWidgets.size() << "]";

            errMsg = errStream.str();
         }
         else
         {
            WidgetStack *ws = env->theWidgets.pop();

            if(ws->size() != 1)
            {
               ostringstream errStream;

               errStream << "Expected one top-level widget, got: ["
                         << env->theWidgets.size() << "]";

               errMsg = errStream.str();
            }
            else
            {
               w = ws->pop();
            }

            delete ws;
         }
      }
   }

   return(w);
}

static void nodeDelete(void *ptr)
{
   Node *n = (Node *)ptr;
   delete n;
}

/*****************************************************************************
 * Globally exported functions
 *****************************************************************************/

char *qale_slurpFile(const char *fileName, string &errMsg)
{
   ostringstream o;
   FILE *f = 0;

   if((f = fopen(fileName, "rb")) == 0)
   {
      o << "Unable to open input file: [" << fileName << "]";

      errMsg = o.str();

      return(0);
   }

   fseek(f, 0, 2);
   int size = ftell(f);
   fseek(f, 0, 0);

   char *text = new char[size + 1];
   int count  = fread(text, size, 1, f);
   text[size] = '\0';
   fclose(f);

   if(count != 1)
   {
      ostringstream errStream;

      errStream << "Error reading: [" << size <<
         "] bytes from file: [" << fileName << "]";

      errMsg = errStream.str();
      return(0);
   }

   return(text);
}

Widget *qale_createFromText(
   const char *text,
   string &errMsg,
   int *errorLineNumberPointer
)
{
   QALEParse *env = new QALEParse();

   Widget *w = qale_createFromText(
      text, env, errMsg, errorLineNumberPointer
   );

   delete env;

   return(w);
}
